# -*- coding: utf-8 -*-

from .commands import plot
from .emase_utils import hybridize, align, bam2emase, quantify

__author__ = 'Kwangbom "KB" Choi, Ph.D.'
__email__ = 'kb.choi@jax.org'
__version__ = '0.1.5'



import os
import sys
import numpy as np
from itertools import combinations_with_replacement
from collections import defaultdict, OrderedDict
from scipy.interpolate import interp1d
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as pyplot


DATA_DIR = os.getenv('GBRS_DATA', '.')


def get_chromosome_info(caller='gbrs'):
    faifile = os.path.join(DATA_DIR, 'ref.fa.fai')
    try:
        chrlens = OrderedDict(np.loadtxt(faifile, usecols=(0, 1), dtype='|S8,<i4'))
    except:
        print >> sys.stderr, "[%s] Make sure if $GBRS_DATA is set correctly. Currently it is %s" % (caller, DATA_DIR)
        raise
    else:
        return chrlens


def get_founder_info(caller='gbrs'):
    fcofile = os.path.join(DATA_DIR, 'founder.hexcolor.info')
    try:
        fcolors = OrderedDict(np.loadtxt(fcofile, usecols=(0, 1), dtype='string', delimiter='\t', comments=None))
    except:
        print >> sys.stderr, "[%s] Make sure if $GBRS_DATA is set correctly. Currently it is %s" % (caller, DATA_DIR)
        raise
    else:
        return fcolors


def unit_vector(vector):
    if sum(vector) > 1e-6:
        return vector / np.linalg.norm(vector)
    else:
        return vector


def get_genotype_probability(aln_profile, aln_specificity, sigma=0.12):
    # 'aln_specificity' should be a set of unit vectors (at least one of the entry is larger than 1.)
    num_haps = len(aln_profile)
    aln_vec = unit_vector(aln_profile)
    genoprob = []
    for i in xrange(num_haps):
        v1 = unit_vector(aln_specificity[i])
        for j in xrange(i, num_haps):
            if j == i:
                genoprob.append(sum(np.power(aln_vec - v1, 2))) # homozygotes
            else:
                v2 = unit_vector(aln_specificity[j])
                geno_vec = unit_vector(v1 + v2)
                # compute directional similarity
                genoprob.append(sum(np.power(aln_vec - geno_vec, 2))) # for heterozygotes
    genoprob = np.exp(np.array(genoprob) / (-2 * sigma * sigma))
    return np.array(genoprob / sum(genoprob))


def reconstruct(**kwargs):
    chrlens = get_chromosome_info(caller='gbrs::reconstruct')
    chrs = chrlens.keys()

    outbase = kwargs.get('outbase')
    if outbase is None:
        out_gtype = 'gbrs.reconstructed.genotypes.tsv'
        out_gprob = 'gbrs.reconstructed.genoprobs.npz'
    else:
        out_gtype = outbase + '.genotypes.tsv'
        out_gprob = outbase + '.genoprobs.npz'
    out_gtype_ordered = os.path.splitext(out_gtype)[0] + '.npz'

    tprobfile = kwargs.get('tprobfile')
    exprfile = kwargs.get('exprfile')
    expr_threshold = kwargs.get('expr_threshold')
    sigma = kwargs.get('sigma')

    # Load alignment specificity
    avecfile = kwargs.get('avecfile')
    if avecfile is None:
        avecfile = os.path.join(DATA_DIR, 'avecs.npz')
    avecs = np.load(avecfile)

    # Load meta info
    gposfile = kwargs.get('gposfile')
    if gposfile is None:
        gposfile = os.path.join(DATA_DIR, 'ref.gene_pos.ordered.npz')
    gene_pos = np.load(gposfile)
    gid_genome_order = dict.fromkeys(gene_pos.files)
    for c in gene_pos.files:
        gid_genome_order[c] = np.array([g for g, p in gene_pos[c]])

    # Load expression level
    expr = dict()
    with open(exprfile) as fh:
        curline = fh.next()
        haplotypes = curline.rstrip().split("\t")[1:-1]
        num_haps = len(haplotypes)
        genotypes = [h1+h2 for h1, h2 in combinations_with_replacement(haplotypes, 2)]
        num_genos = len(genotypes)
        for curline in fh:
            item = curline.rstrip().split("\t")
            expr[item[0]] = np.array(map(float, item[1:-1]))

    # Get null model probability
    init_vec = []
    for h1, h2 in combinations_with_replacement(haplotypes, 2):
        if h1 == h2:
            init_vec.append(np.log(1.0/(num_haps*num_haps)))
        else:
            init_vec.append(np.log(2.0/(num_haps*num_haps)))
    init_vec = np.array(init_vec)

    # Get initial emission probability
    naiv_avecs = np.eye(num_haps) + (np.ones((num_haps, num_haps)) - np.eye(num_haps)) * 0.0001
    eprob = dict()
    for gid, evec in expr.iteritems():
        if sum(evec) < expr_threshold:
            eprob[gid] = init_vec
        elif gid not in avecs.files:
            eprob[gid] = np.log(get_genotype_probability(evec, naiv_avecs, sigma=0.450) + np.nextafter(0, 1))
        else:
            eprob[gid] = np.log(get_genotype_probability(evec, avecs[gid], sigma=sigma) + np.nextafter(0, 1))

    # Load transition probabilities
    tprob = np.load(tprobfile)

    # Get forward probability
    alpha = dict()
    alpha_scaler = dict()
    for c in chrs:
        if c in tprob.files:
            tprob_c = tprob[c]
            gid_genome_order_c = gid_genome_order[c]
            num_genes_in_chr = len(gid_genome_order_c)
            alpha_c = np.zeros((num_genos, num_genes_in_chr))
            alpha_scaler_c = np.zeros(num_genes_in_chr)
            alpha_c[:, 0] = init_vec + eprob[gid_genome_order_c[0]]
            normalizer = np.log(sum(np.exp(alpha_c[:, 0])))
            alpha_c[:, 0] -= normalizer # normalization
            alpha_scaler_c[0] = -normalizer
            for i in xrange(1, num_genes_in_chr):
                alpha_c[:, i] = np.log(np.exp(alpha_c[:, i-1] + tprob_c[i-1]).sum(axis=1) + np.nextafter(0, 1)) + \
                                eprob[gid_genome_order_c[i]]
                normalizer = np.log(sum(np.exp(alpha_c[:, i])))
                alpha_c[:, i] -= normalizer  # normalization
                alpha_scaler_c[i] = -normalizer
            alpha[c] = alpha_c
            alpha_scaler[c] = alpha_scaler_c

    # Get backward probability
    beta = dict()
    for c in chrs:
        if c in tprob.files:
            tprob_c = tprob[c]
            gid_genome_order_c = gid_genome_order[c]
            num_genes_in_chr = len(gid_genome_order_c)
            beta_c = np.zeros((num_genos, num_genes_in_chr))
            beta_c[:, -1] = alpha_scaler[c][-1]  #init_vec + eprob[gid_genome_order_c[-1]]
            for i in xrange(num_genes_in_chr-2, -1, -1):
                beta_c[:, i] = np.log(np.exp(tprob_c[i].transpose() +
                                             beta_c[:, i+1] +
                                             eprob[gid_genome_order_c[i+1]] +
                                             alpha_scaler[c][i]).sum(axis=1))
            beta[c] = beta_c

    # Get forward-backward probability
    gamma = dict()
    for c in chrs:
        if c in tprob.files:
            gamma_c = np.exp(alpha[c] + beta[c])
            normalizer = gamma_c.sum(axis=0)
            gamma[c] = gamma_c / normalizer
    np.savez_compressed(out_gprob, **gamma)

    # Run Viterbi
    delta = dict()
    for c in chrs:
        if c in tprob.files:
            tprob_c = tprob[c]
            gid_genome_order_c = gid_genome_order[c]
            num_genes_in_chr = len(gid_genome_order_c)
            delta_c = np.zeros((num_genos, num_genes_in_chr))
            delta_c[:, 0] = init_vec + eprob[gid_genome_order_c[0]]
            for i in xrange(1, num_genes_in_chr):
                delta_c[:, i] = (delta_c[:, i-1] + tprob_c[i-1]).max(axis=1) + eprob[gid_genome_order_c[i]]
            delta[c] = delta_c
    viterbi_states = defaultdict(list)
    gtcall_g = dict()
    for c in chrs:
        if c in tprob.files:
            tprob_c = tprob[c]
            gid_genome_order_c = gid_genome_order[c]
            num_genes_in_chr = len(gid_genome_order_c)
            sid = delta[c][:, num_genes_in_chr-1].argmax()
            viterbi_states[c].append(genotypes[sid])
            for i in reversed(xrange(num_genes_in_chr-1)):
                sid = (delta[c][:, i] + tprob_c[i][sid]).argmax()
                viterbi_states[c].append(genotypes[sid])
                gtcall_g[gid_genome_order_c[i]] = genotypes[sid]
            viterbi_states[c].reverse()
    viterbi_states = dict(viterbi_states)
    np.savez_compressed(out_gtype_ordered, **viterbi_states)
    with open(out_gtype, 'w') as fhout:
        fhout.write('#Gene_ID\tDiplotype\n')
        for g in sorted(gtcall_g.keys()):
            fhout.write('%s\t%s\n' % (g, gtcall_g[g]))


def interpolate(**kwargs):
    chrlens = get_chromosome_info(caller='gbrs::interpolate')
    chrs = chrlens.keys()

    probfile = kwargs.get('probfile')

    gposfile = kwargs.get('gposfile')
    if gposfile is None:  # if gposfile is not specified
        gposfile = os.path.join(DATA_DIR, 'ref.gene_pos.ordered.npz')
        try:
            x_gene = np.load(gposfile)
        except:
            print >> sys.stderr, "[gbrs::interpolate] Please make sure if $GBRS_DATA is set correctly: %s" % DATA_DIR
            raise
        else:
            pass
    else:
        x_gene = np.load(gposfile)

    gridfile = kwargs.get('gridfile')
    if gridfile is None:
        gridfile = os.path.join(DATA_DIR, 'ref.genome_grid.64k.txt')

    outfile = kwargs.get('outfile')
    if outfile is None:
        outfile = 'gbrs.interpolated.' + os.path.basename(probfile)

    x_grid = defaultdict(list)
    with open(gridfile) as fh:
        fh.next()  # skip header
        for curline in fh:
            item = curline.rstrip().split("\t")
            x_grid[item[0]].append(float(item[1]))
    x_grid = dict(x_grid)

    x_grid_complete = dict()
    for c in chrs:
        if c in x_gene.files:
            x = [float(coord) for m, coord in x_gene[c]]
            x_min = min(x_grid[c][0]-100.0, 0.0)
            x_max = max(x_grid[c][-1]+100.0, chrlens[c])
            x = np.append([x_min], x)
            x = np.append(x, [x_max])
            x_grid_complete[c] = x

    gamma_gene = np.load(probfile)
    gene_model_chr = dict()
    gene_intrp_chr = dict()
    for c in chrs:
        if c in gamma_gene.files:
            gamma_gene_c = gamma_gene[c]
            y = np.hstack((gamma_gene_c[:, 0][:, np.newaxis], gamma_gene_c))
            y = np.hstack((y, y[:, -1][:, np.newaxis]))
            gene_model_chr[c] = interp1d(x_grid_complete[c], y, axis=1)
            gene_intrp_chr[c] = gene_model_chr[c](x_grid[c])
    np.savez_compressed(outfile, **gene_intrp_chr)


def combine(**kwargs):
    raise NotImplementedError


def plot(**kwargs):
    chrlens = get_chromosome_info(caller='gbrs::plot')
    chrs = chrlens.keys()
    num_chrs = len(chrs)

    gpbfile = kwargs.get('gpbfile')
    outfile = kwargs.get('outfile')
    if outfile is None:
        outfile = 'gbrs.plotted.' + os.path.splitext(os.path.basename(gpbfile))[0] + '.pdf'

    sample_name = kwargs.get('sample_name')
    grid_size = kwargs.get('grid_size')
    xt_max = kwargs.get('xt_max')
    xt_size = kwargs.get('xt_size')
    width = kwargs.get('width')

    hcolors = get_founder_info(caller='gbrs::plot')
    haplotypes = hcolors.keys()
    hid = dict(zip(haplotypes, np.arange(8)))
    genotypes = np.array([h1+h2 for h1, h2 in combinations_with_replacement(haplotypes, 2)])

    #
    # Main body
    #
    genoprob = np.load(gpbfile)
    fig = pyplot.figure()
    fig.set_size_inches((16, 16))
    ax = fig.add_subplot(111)
    ax.set_xlim(0, 4500*width+width)
    ax.set_ylim(1, 95)
    num_recomb_total = 0
    for cid, c in enumerate(chrs):
        if c in genoprob.files:  # Skip drawing Y chromosome if the sample is female
            genotype_calls = genotypes[genoprob[c].argmax(axis=0)]
            hap = []
            col1 = []
            col2 = []
            oldcol1 = 'NA'
            oldcol2 = 'NA'
            num_recomb = 0
            num_genes_in_chr = len(genotype_calls)
            for i in xrange(num_genes_in_chr):
                hap.append((i*width, width))
                c1 = hcolors[genotype_calls[i][0]]
                c2 = hcolors[genotype_calls[i][1]]
                if i > 0:
                    if c1 == c2:
                        if col1[-1] != col2[-1]: # When homozygous region starts, remember the most recent het
                            oldcol1 = col1[-1]
                            oldcol2 = col2[-1]
                    else:
                        if col1[-1] == col2[-1]: # When heterozygous region starts
                            if c1 == oldcol2 or c2 == oldcol1:
                                c1, c2 = c2, c1
                        elif c1 == col2[-1] or c2 == col1[-1]:
                            c1, c2 = c2, c1
                    if c1 != col1[-1] or c2 != col2[-1]:
                        num_recomb += 1
                col1.append(c1)
                col2.append(c2)
            num_recomb_total += num_recomb
            # plot
            ax.broken_barh(hap, (num_chrs*4-cid*4+1, 1), facecolors=col1, edgecolor='face')
            ax.broken_barh(hap, (num_chrs*4-cid*4, 1), facecolors=col2, edgecolor='face')
            ax.text((num_genes_in_chr+50)*width, num_chrs*4-cid*4+0.5, '(%d)' % num_recomb)
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.get_xaxis().tick_bottom()
        ax.get_yaxis().tick_right()
        ax.get_yaxis().tick_left()
        ax.get_yaxis().set_ticks([])
        ax.set_yticklabels(chrs)
        pyplot.yticks(np.arange(num_chrs*4+1, 1, -4), fontsize=14)
        ax.set_xticklabels([ '%dM' % xt for xt in np.arange(0, xt_max*grid_size/1000000, xt_size*grid_size/1000000)])
        pyplot.xticks(np.arange(0, xt_max*width, xt_size*width))
        title_txt = 'Genome reconstruction: ' + sample_name
        title_txt += "\n(Total %d recombinations)" % num_recomb_total
        ax.set_title(title_txt, fontsize=18, loc='center')
    fig.savefig(outfile, dpi=300)
    pyplot.close(fig)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import gbrs
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = []
on_rtd = os.environ.get('READTHEDOCS', None)
if not on_rtd:
    # requirements.append('numpy==1.8.2')
    requirements.append('g2gtools')
    requirements.append('emase')
    requirements.append('matplotlib')
    requirements.append('bx-python>=0.7.2')
    requirements.append('pysam>=0.8.1')
    requirements.append('biopython>=1.63')
    requirements.append('pysqlite>=2.6.3')
    requirements.append('tables>=3.1.0')

test_requirements = [
    'pytest'
]

setup(
    name='gbrs',
    version=gbrs.__version__,
    description='A suite of tools for Reconstructing Genomes and Quantifying Allele Specific Expression from RNA-Seq data',
    long_description=readme + '\n\n' + history,
    author='Kwangbom \"KB\" Choi, Ph.D.',
    author_email='kb.choi@jax.org',
    url='https://github.com/churchill-lab/gbrs',
    packages=[
        'gbrs',
    ],
    package_dir={'gbrs':
                 'gbrs'},
    include_package_data=True,
    install_requires=requirements,
    license="GPLv3",
    zip_safe=False,
    keywords='gbrs',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
    ],
    scripts=['scripts/gbrs',
             'scripts/convert-kallisto-result-for-gbrs',
             'scripts/convert-salmon-result-for-gbrs'],
    test_suite='tests',
    tests_require=test_requirements
)



import os
import sys
import numpy as np
import subprocess
from itertools import dropwhile
from collections import defaultdict
import emase


DATA_DIR = os.getenv('GBRS_DATA', '.')


def is_comment(s):
    return s.startswith('#')


def get_names(idfile):
    ids = dict()
    master_id = 0
    with open(idfile) as fh:
        for curline in fh:
            item = curline.rstrip().split("\t")
            g = item[0]
            if not ids.has_key(g):
                ids[g] = master_id
                master_id += 1
    num_ids = len(ids)
    names = {index:name for name, index in ids.iteritems()}
    return [names[k] for k in xrange(num_ids)]


def hybridize(**kwargs):
    outfile = kwargs.get('outfile')
    outdir = os.path.dirname(outfile)
    if not os.path.exists(outdir):
        os.mkdir(outdir)

    # Get pooled transcriptome
    outbase = os.path.splitext(outfile)[0]
    fastalist = kwargs.get('fastalist').split(',')
    haplist = kwargs.get('haplist').split(',')
    num_haps = len(fastalist)
    lenfile = outbase + '.info'
    seqout = open(outfile, 'w')
    lenout = open(lenfile, 'w')
    for hid in xrange(num_haps):
        fasta = fastalist[hid]
        hapname = haplist[hid]
        print >> sys.stderr, "[gbrs::hybridize] Adding suffix \'_%s\' to the sequence ID's of %s..." % (hapname, fasta)
        fh = open(fasta)
        curline = fh.next()  # The first fasta header
        curline = curline.rstrip().split()[0] + '_' + hapname
        seqout.write('%s\n' % curline)
        lenout.write('%s\t' % curline[1:])
        seqlen = 0
        for curline in fh:
            if curline[0] == '>':
                curline = curline.rstrip().split()[0] + '_' + hapname + "\n"
                lenout.write('%d\n%s\t' % (seqlen, curline[1:].rstrip()))
                seqlen = 0
            else:
                seqlen += len(curline.rstrip())
            seqout.write(curline)
        fh.close()
        lenout.write('%d\n' % seqlen)
    seqout.close()
    lenout.close()

    # Build bowtie index for the pooled transcriptome
    build_bowtie_index = kwargs.get('build_bowtie_index')
    if build_bowtie_index:
        out_index = outbase + '.bowtie1'
        print >> sys.stderr, "[gbrs::hybridize] Building bowtie1 index..."
        status = subprocess.call("bowtie-build %s %s" % (outfile, out_index), shell=True)


def align(**kwargs):
    raise NotImplementedError


def bam2emase(**kwargs):
    alnfile = kwargs.get('alnfile')
    lidfile = kwargs.get('lidfile')
    if lidfile is None:
        lidfile2chk = os.path.join(DATA_DIR, 'ref.transcripts.info')
        if os.path.exists(lidfile2chk):
            lidfile = lidfile2chk
        else:
            print >> sys.stderr, '[gbrs::bam2emase] Cannot find a locus id file.'
            return 2
    outfile = kwargs.get('outfile')
    if outfile is None:
        outfile = 'gbrs.bam2emased.' + os.path.splitext(os.path.basename(alnfile))[0] + '.h5'

    haplotypes = tuple(kwargs.get('haplogypes').split(','))
    index_dtype = kwargs.get('index_dtype')
    data_dtype = kwargs.get('data_dtype')

    loci = get_names(lidfile)
    alignmat_factory = emase.AlignmentMatrixFactory(alnfile)
    alignmat_factory.prepare(haplotypes, loci, outdir=os.path.dirname(outfile))
    alignmat_factory.produce(outfile, index_dtype=index_dtype, data_dtype=data_dtype)
    alignmat_factory.cleanup()


def intersect(**kwargs):
    flist = kwargs.get('emasefiles').split(',')
    outfile = kwargs.get('outfile')
    if outfile is None:
        outfile = 'gbrs.intersected.' + os.path.basename(flist[0])
    complib = kwargs.get('complib')
    alnmat = emase.AlignmentPropertyMatrix(h5file=flist[0])
    for f in flist[1:]:
        alnmat_next = emase.AlignmentPropertyMatrix(h5file=f)
        if np.all(alnmat.rname == alnmat_next.rname):
            alnmat = alnmat * alnmat_next
        else:
            print >> sys.stderr, "[gbrs::intersect] The read ID's are not compatible."
            return 2
    alnmat.save(h5file=outfile, complib=complib)


def compress(**kwargs):
    flist = kwargs.get('emasefiles').split(',')
    outfile = kwargs.get('outfile')
    complib = kwargs.get('complib')

    ec = defaultdict(int)
    for alnfile in flist:
        alnmat_rd = emase.AlignmentPropertyMatrix(h5file=alnfile)
        for h in xrange(alnmat_rd.num_haplotypes):
            alnmat_rd.data[h] = alnmat_rd.data[h].tocsr()
        if alnmat_rd.count is None:
            alnmat_rd.count = np.ones(alnmat_rd.num_reads)
        for curind in xrange(alnmat_rd.num_reads):
            ec_key = []
            for h in xrange(alnmat_rd.num_haplotypes):
                i0 = alnmat_rd.data[h].indptr[curind]
                i1 = alnmat_rd.data[h].indptr[curind+1]
                ec_key.append(','.join(map(str, sorted(alnmat_rd.data[h].indices[i0:i1]))))
            ec[':'.join(ec_key)] += alnmat_rd.count[curind]
    ec = dict(ec)
    num_ecs = len(ec)
    alnmat_ec = emase.AlignmentPropertyMatrix(shape=(alnmat_rd.num_loci, alnmat_rd.num_haplotypes, num_ecs))
    alnmat_ec.hname = alnmat_rd.hname
    alnmat_ec.lname = alnmat_rd.lname
    alnmat_ec.count = np.zeros(num_ecs)
    for row_id, ec_key in enumerate(ec):
        alnmat_ec.count[row_id] = ec[ec_key]
        nzlocs = ec_key.split(':')
        for h in xrange(alnmat_ec.num_haplotypes):
            nzlocs_h = nzlocs[h]
            if nzlocs_h != '':
                nzinds = np.array(map(np.int, nzlocs_h.split(',')))
                alnmat_ec.data[h][row_id, nzinds] = 1
    alnmat_ec.finalize()
    alnmat_ec.save(h5file=outfile, complib=complib)


def stencil(**kwargs):
    """
    Applying genotype calls to multi-way alignment incidence matrix

    :param alnfile: alignment incidence file (h5),
    :param gtypefile: genotype calls by GBRS (tsv),
    :param grpfile: gene ID to isoform ID mapping info (tsv)
    :return: genotyped version of alignment incidence file (h5)
    """
    alnfile = kwargs.get('alnfile')
    gtypefile = kwargs.get('gtypefile')
    grpfile = kwargs.get('grpfile')
    if grpfile is None:
        grpfile2chk = os.path.join(DATA_DIR, 'ref.gene2transcripts.tsv')
        if os.path.exists(grpfile2chk):
            grpfile = grpfile2chk
        else:
            print >> sys.stderr, '[gbrs::stencil] A group file is *not* given. Genotype will be stenciled as is.'

    # Load alignment incidence matrix ('alnfile' is assumed to be in multiway transcriptome)
    alnmat = emase.AlignmentPropertyMatrix(h5file=alnfile, grpfile=grpfile)

    # Load genotype calls
    hid = dict(zip(alnmat.hname, np.arange(alnmat.num_haplotypes)))
    gid = dict(zip(alnmat.gname, np.arange(len(alnmat.gname))))
    gtmask = np.zeros((alnmat.num_haplotypes, alnmat.num_loci))
    gtcall_g = dict.fromkeys(alnmat.gname)
    with open(gtypefile) as fh:
        if grpfile is not None:
            gtcall_t = dict.fromkeys(alnmat.lname)
            for curline in dropwhile(is_comment, fh):
                item = curline.rstrip().split("\t")
                g, gt = item[:2]
                gtcall_g[g] = gt
                hid2set = np.array([hid[c] for c in gt])
                tid2set = np.array(alnmat.groups[gid[g]])
                gtmask[np.meshgrid(hid2set, tid2set)] = 1.0
                for t in tid2set:
                    gtcall_t[alnmat.lname[t]] = gt
        else:
            for curline in dropwhile(is_comment, fh):
                item = curline.rstrip().split("\t")
                g, gt = item[:2]
                gtcall_g[g] = gt
                hid2set = np.array([hid[c] for c in gt])
                gtmask[np.meshgrid(hid2set, gid[g])] = 1.0

    alnmat.multiply(gtmask, axis=2)
    for h in xrange(alnmat.num_haplotypes):
        alnmat.data[h].eliminate_zeros()
    outfile = kwargs.get('outfile')
    if outfile is None:
        outfile = 'gbrs.stenciled.' + os.path.basename(alnfile)
    alnmat.save(h5file=outfile)


def quantify(**kwargs):
    """
    Quantify expected read counts

    :param alnfile: alignment incidence file (h5)
    :param grpfile: gene ID to isoform ID mapping info (tsv)
    :param lenfile: transcript lengths (tsv)
    :param multiread_model: emase model (default: 4)
    :param read_length: read length (default: 100)
    :param pseudocount: prior read count (default: 0.0)
    :param tolerance: tolerance for EM termination (default: 0.0001 in TPM)
    :param max_iters: maximum iterations for EM iteration
    :param report_alignment_counts: whether to report alignment counts (default: False)
    :param report_posterior:
    :return: Expected read counts (tsv)
    """
    alnfile = kwargs.get('alnfile')
    grpfile = kwargs.get('grpfile')
    if grpfile is None:
        grpfile2chk = os.path.join(DATA_DIR, 'ref.gene2transcripts.tsv')
        if os.path.exists(grpfile2chk):
            grpfile = grpfile2chk
        else:
            print >> sys.stderr, '[gbrs::quantify] A group file is not given. Group-level results will not be reported.'

    outbase = kwargs.get('outbase')
    gtypefile = kwargs.get('gtypefile')
    pseudocount = kwargs.get('pseudocount')
    lenfile = kwargs.get('lenfile')
    if lenfile is None:
        lenfile2chk = os.path.join(DATA_DIR, 'gbrs.hybridized.targets.info')
        if os.path.exists(lenfile2chk):
            lenfile = lenfile2chk
        else:
            print >> sys.stderr, '[gbrs::quantify] A length file is not given. Transcript length adjustment will *not* be performed.'

    read_length = kwargs.get('read_length')
    multiread_model = kwargs.get('multiread_model')
    tolerance = kwargs.get('tolerance')
    max_iters = kwargs.get('max_iters')
    report_group_counts = grpfile is not None  # If grpfile exist, always report groupwise results too
    report_alignment_counts = kwargs.get('report_alignment_counts')
    report_posterior = kwargs.get('report_posterior')

    # Load alignment incidence matrix ('alnfile' is assumed to be in multiway transcriptome)
    alnmat = emase.AlignmentPropertyMatrix(h5file=alnfile, grpfile=grpfile)

    # Load genotype calls
    if gtypefile is not None:  # Genotype calls are at the gene level
        outbase = outbase + '.diploid'
        hid = dict(zip(alnmat.hname, np.arange(alnmat.num_haplotypes)))
        gid = dict(zip(alnmat.gname, np.arange(len(alnmat.gname))))
        gtmask = np.zeros((alnmat.num_haplotypes, alnmat.num_loci))
        gtcall_g = dict.fromkeys(alnmat.gname)
        gtcall_t = dict.fromkeys(alnmat.lname)
        with open(gtypefile) as fh:
            for curline in dropwhile(is_comment, fh):
                item = curline.rstrip().split("\t")
                g, gt = item[:2]
                gtcall_g[g] = gt
                hid2set = np.array([hid[c] for c in gt])
                tid2set = np.array(alnmat.groups[gid[g]])
                gtmask[np.meshgrid(hid2set, tid2set)] = 1.0
                for t in tid2set:
                    gtcall_t[alnmat.lname[t]] = gt
        alnmat.multiply(gtmask, axis=2)
        for h in xrange(alnmat.num_haplotypes):
            alnmat.data[h].eliminate_zeros()
    else:
        outbase = outbase + ".multiway"
        gtcall_g = None
        gtcall_t = None

    # Run emase
    em_factory = emase.EMfactory(alnmat)
    em_factory.prepare(pseudocount=pseudocount, lenfile=lenfile, read_length=read_length)
    em_factory.run(model=multiread_model, tol=tolerance, max_iters=max_iters, verbose=True)
    em_factory.report_depths(filename="%s.isoforms.tpm" % outbase, tpm=True, notes=gtcall_t)
    em_factory.report_read_counts(filename="%s.isoforms.expected_read_counts" % outbase, notes=gtcall_t)
    if report_posterior:
        em_factory.export_posterior_probability(filename="%s.posterior.h5" % outbase)
    if report_group_counts:
        em_factory.report_depths(filename="%s.genes.tpm" % outbase, tpm=True, grp_wise=True, notes=gtcall_g)
        em_factory.report_read_counts(filename="%s.genes.expected_read_counts" % outbase, grp_wise=True, notes=gtcall_g)
    if report_alignment_counts:
        alnmat = emase.AlignmentPropertyMatrix(h5file=alnfile, grpfile=grpfile)
        alnmat.report_alignment_counts(filename="%s.isoforms.alignment_counts" % outbase)
        if report_group_counts:
            alnmat._bundle_inline(reset=True)
            alnmat.report_alignment_counts(filename="%s.genes.alignment_counts" % outbase)
